@@ -41,26 +41,26 @@ let make_name_from_path { Url.Path.name; parent; _ } =
   | None -> name
   | Some p -> Printf.sprintf "%s.%s" p.name name
 
-let label t ppf =
+let label t =
   match t with
-  | Odoc_model.Lang.TypeExpr.Label s -> O.pf ppf "%s" s
-  | Optional s -> O.pf ppf "?%s" s
+  | Odoc_model.Lang.TypeExpr.Label s -> O.txt s
+  | Optional s -> O.txt "?" ++ O.txt s
 
-let tag tag t ppf = O.pf ppf "@{<%s>%t@}" tag t
+let tag tag t = O.span ~attr:tag t
 
 let type_var tv = tag "type-var" (O.txt tv)
 
-let enclose ~l ~r x = O.span (fun ppf -> O.pf ppf "%s%t%s" l x r)
+let enclose ~l ~r x = O.span (O.txt l ++ x ++ O.txt r)
 
 let path p txt =
-  !O.elt
+  O.elt
     [ inline @@ InternalLink (InternalLink.Resolved (Url.from_path p, txt)) ]
 
 let resolved p txt =
-  !O.elt [ inline @@ InternalLink (InternalLink.Resolved (p, txt)) ]
+  O.elt [ inline @@ InternalLink (InternalLink.Resolved (p, txt)) ]
 
 let unresolved txt =
-  !O.elt [ inline @@ InternalLink (InternalLink.Unresolved txt) ]
+  O.elt [ inline @@ InternalLink (InternalLink.Unresolved txt) ]
 
 let path_to_id path =
   match Url.Anchor.from_identifier (path :> Paths.Identifier.t) with
@@ -235,9 +235,8 @@ module Make (Syntax : SYNTAX) = struct
             let first =
               match first with
               | Odoc_model.Lang.TypeExpr.Polymorphic_variant.Type te ->
-                  let res = type_expr te in
-                  if add_pipe then O.txt " " ++ O.span (O.txt "| " ++ res)
-                  else res
+                  let res = O.box_hv @@ type_expr te in
+                  if add_pipe then O.sp ++ O.span (O.txt "| " ++ res) else res
               | Constructor { constant; name; arguments; _ } ->
                   let constr =
                     let name = "`" ^ name in
@@ -245,37 +244,42 @@ module Make (Syntax : SYNTAX) = struct
                     else O.txt name
                   in
                   let res =
-                    match arguments with
-                    | [] -> constr
-                    | _ ->
-                        let arguments = style_arguments ~constant arguments in
-                        O.span
-                          (if Syntax.Type.Variant.parenthesize_params then
-                           constr ++ arguments
-                          else constr ++ O.txt " of " ++ arguments)
+                    O.box_hv
+                      (match arguments with
+                      | [] -> constr
+                      | _ ->
+                          let arguments = style_arguments ~constant arguments in
+                          O.span
+                            (if Syntax.Type.Variant.parenthesize_params then
+                             constr ++ arguments
+                            else constr ++ O.txt " of" ++ O.sp ++ arguments))
                   in
-                  if add_pipe then O.txt " " ++ res else res
+                  if add_pipe then O.sp ++ res else res
             in
             first ++ style_elements ~add_pipe:true rest
       in
       let elements = style_elements ~add_pipe:false t.elements in
-      O.span
-        (match t.kind with
-        | Fixed -> O.txt "[ " ++ elements ++ O.txt " ]"
-        | Open -> O.txt "[> " ++ elements ++ O.txt " ]"
-        | Closed [] -> O.txt "[< " ++ elements ++ O.txt " ]"
-        | Closed lst ->
-            let constrs = String.concat " " lst in
-            O.txt "[< " ++ elements ++ O.txt (" " ^ constrs ^ " ]"))
+      O.box_hv_no_indent
+      @@ O.span
+           (match t.kind with
+           | Fixed -> O.txt "[ " ++ elements ++ O.txt " ]"
+           | Open -> O.txt "[> " ++ elements ++ O.txt " ]"
+           | Closed [] -> O.txt "[< " ++ elements ++ O.txt " ]"
+           | Closed lst ->
+               let constrs = String.concat " " lst in
+               O.txt "[< " ++ elements ++ O.txt (" " ^ constrs ^ " ]"))
 
     and te_object (t : Odoc_model.Lang.TypeExpr.Object.t) =
       let fields =
-        O.list t.fields ~f:(function
-          | Odoc_model.Lang.TypeExpr.Object.Method { name; type_ } ->
-              O.txt (name ^ Syntax.Type.annotation_separator)
-              ++ type_expr type_
-              ++ O.txt Syntax.Obj.field_separator
-          | Inherit type_ -> type_expr type_ ++ O.txt Syntax.Obj.field_separator)
+        O.list
+          ~sep:(O.sp ++ O.txt Syntax.Obj.field_separator)
+          t.fields
+          ~f:(function
+            | Odoc_model.Lang.TypeExpr.Object.Method { name; type_ } ->
+                O.box_hv_no_indent
+                @@ O.txt (name ^ Syntax.Type.annotation_separator)
+                   ++ O.cut ++ type_expr type_
+            | Inherit type_ -> O.box_hv_no_indent @@ type_expr type_)
       in
       let open_tag =
         if t.open_ then O.txt Syntax.Obj.open_tag_extendable
@@ -289,6 +293,8 @@ module Make (Syntax : SYNTAX) = struct
 
     and format_type_path ~delim (params : Odoc_model.Lang.TypeExpr.t list)
         (path : text) : text =
+      O.box_hv
+      @@
       match params with
       | [] -> path
       | [ param ] ->
@@ -300,13 +306,13 @@ module Make (Syntax : SYNTAX) = struct
           in
           Syntax.Type.handle_constructor_params path args
       | params ->
-          let params = O.list params ~sep:(O.txt ",\194\160") ~f:type_expr in
+          let params = O.list params ~sep:(O.txt "," ++ O.sp) ~f:type_expr in
           let params =
             match delim with
-            | `parens -> enclose ~l:"(" params ~r:")"
-            | `brackets -> enclose ~l:"[" params ~r:"]"
+            | `parens -> enclose ~l:"( " params ~r:" )"
+            | `brackets -> enclose ~l:"[ " params ~r:" ]"
           in
-          Syntax.Type.handle_constructor_params path params
+          Syntax.Type.handle_constructor_params path (O.box_hv params)
 
     and type_expr ?(needs_parentheses = false) (t : Odoc_model.Lang.TypeExpr.t)
         =
@@ -319,25 +325,27 @@ module Make (Syntax : SYNTAX) = struct
       | Arrow (None, src, dst) ->
           let res =
             O.span
-              (type_expr ~needs_parentheses:true src
+              ((O.box_hv @@ type_expr ~needs_parentheses:true src)
               ++ O.txt " " ++ Syntax.Type.arrow)
-            ++ O.txt " " ++ type_expr dst
+            ++ O.sp ++ type_expr dst
+            (* ++ O.end_hv *)
           in
-          if not needs_parentheses then res else enclose ~l:"(" res ~r:")"
+          if not needs_parentheses then res else enclose ~l:"( " res ~r:" )"
       | Arrow (Some lbl, src, dst) ->
           let res =
             O.span
-              (label lbl ++ O.txt ":"
-              ++ type_expr ~needs_parentheses:true src
-              ++ O.sp ++ Syntax.Type.arrow)
+              ((O.box_hv
+               @@ label lbl ++ O.txt ":" ++ O.cut
+                  ++ (O.box_hv @@ type_expr ~needs_parentheses:true src))
+              ++ O.txt " " ++ Syntax.Type.arrow)
             ++ O.sp ++ type_expr dst
           in
-          if not needs_parentheses then res else enclose ~l:"(" res ~r:")"
+          if not needs_parentheses then res else enclose ~l:"( " res ~r:" )"
       | Tuple lst ->
           let res =
-            O.list lst
-              ~sep:(O.txt Syntax.Type.Tuple.element_separator)
-              ~f:(type_expr ~needs_parentheses:true)
+            O.box_hv_no_indent
+              (O.list lst ~sep:Syntax.Type.Tuple.element_separator
+                 ~f:(type_expr ~needs_parentheses:true))
           in
           if Syntax.Type.Tuple.always_parenthesize || needs_parentheses then
             enclose ~l:"(" res ~r:")"
@@ -351,7 +359,7 @@ module Make (Syntax : SYNTAX) = struct
           format_type_path ~delim:`brackets args
             (Link.from_path (path :> Paths.Path.t))
       | Poly (polyvars, t) ->
-          O.txt (String.concat " " polyvars ^ ". ") ++ type_expr t
+          O.txt ("'" ^ String.concat " '" polyvars ^ ". ") ++ type_expr t
       | Package pkg ->
           enclose ~l:"(" ~r:")"
             (O.keyword "module" ++ O.txt " "
@@ -359,17 +367,21 @@ module Make (Syntax : SYNTAX) = struct
             ++
             match pkg.substitutions with
             | [] -> O.noop
-            | lst ->
-                O.txt " " ++ O.keyword "with" ++ O.txt " "
-                ++ O.list
-                     ~sep:(O.txt " " ++ O.keyword "and" ++ O.txt " ")
-                     lst ~f:package_subst)
+            | fst :: lst ->
+                O.sp
+                ++ O.box_hv (O.keyword "with" ++ O.txt " " ++ package_subst fst)
+                ++ O.list lst ~f:(fun s ->
+                       O.cut
+                       ++ (O.box_hv
+                          @@ O.txt " " ++ O.keyword "and" ++ O.txt " "
+                             ++ package_subst s)))
 
     and package_subst
         ((frag_typ, te) : Paths.Fragment.Type.t * Odoc_model.Lang.TypeExpr.t) :
         text =
       let typath = Link.from_fragment (frag_typ :> Paths.Fragment.leaf) in
-      O.keyword "type" ++ O.txt " " ++ typath ++ O.txt " = " ++ type_expr te
+      O.keyword "type" ++ O.txt " " ++ typath ++ O.txt " =" ++ O.sp
+      ++ type_expr te
   end
 
   open Type_expression
@@ -464,8 +476,7 @@ module Make (Syntax : SYNTAX) = struct
       | Tuple [] -> O.documentedSrc (cstr ++ ret_type)
       | Tuple lst ->
           let params =
-            O.list lst
-              ~sep:(O.txt Syntax.Type.Tuple.element_separator)
+            O.list lst ~sep:Syntax.Type.Tuple.element_separator
               ~f:(type_expr ~needs_parentheses:is_gadt)
           in
           O.documentedSrc
@@ -558,11 +569,13 @@ module Make (Syntax : SYNTAX) = struct
           DocumentedSrc.Nested { anchor; attrs; code; doc; markers }
 
     let extension (t : Odoc_model.Lang.Extension.t) =
+      let prefix =
+        O.keyword "type" ++ O.txt " "
+        ++ Link.from_path (t.type_path :> Paths.Path.t)
+        ++ O.txt " +=" ++ O.sp
+      in
       let content =
-        O.documentedSrc
-          (O.keyword "type" ++ O.txt " "
-          ++ Link.from_path (t.type_path :> Paths.Path.t)
-          ++ O.txt " += ")
+        O.documentedSrc prefix
         @ List.map extension_constructor t.constructors
         @ O.documentedSrc
             (if Syntax.Type.type_def_semicolon then O.txt ";" else O.noop)
@@ -610,7 +623,10 @@ module Make (Syntax : SYNTAX) = struct
                       else fun x -> type_expr x
                     in
                     let params =
-                      O.list arguments ~sep:(O.txt " & ") ~f:wrapped_type_expr
+                      O.box_hv
+                      @@ O.list arguments
+                           ~sep:(O.txt " &" ++ O.sp)
+                           ~f:wrapped_type_expr
                     in
                     let params =
                       if constant then O.txt "& " ++ params else params
@@ -619,7 +635,7 @@ module Make (Syntax : SYNTAX) = struct
                       (O.txt cstr
                       ++
                       if Syntax.Type.Variant.parenthesize_params then params
-                      else O.txt " " ++ O.keyword "of" ++ O.txt " " ++ params)),
+                      else O.txt " " ++ O.keyword "of" ++ O.sp ++ params)),
                 match doc with [] -> None | _ -> Some (Comment.to_ir doc) ))
         in
         let markers = Syntax.Comment.markers in
@@ -686,8 +702,12 @@ module Make (Syntax : SYNTAX) = struct
 
     let format_constraints constraints =
       O.list constraints ~f:(fun (t1, t2) ->
-          O.txt " " ++ O.keyword "constraint" ++ O.txt " " ++ type_expr t1
-          ++ O.txt " = " ++ type_expr t2)
+          O.sp
+          ++ (O.box_hv
+             @@ O.keyword "constraint" ++ O.sp
+                ++ O.box_hv_no_indent (type_expr t1)
+                ++ O.txt " =" ++ O.sp
+                ++ O.box_hv_no_indent (type_expr t2)))
 
     let format_manifest :
           'inner_row 'outer_row.
@@ -703,7 +723,8 @@ module Make (Syntax : SYNTAX) = struct
       | None -> (O.noop, private_)
       | Some t ->
           let manifest =
-            O.txt (if is_substitution then " := " else " = ")
+            O.txt (if is_substitution then " :=" else " =")
+            ++ O.sp
             ++ (if private_ then
                 O.keyword Syntax.Type.private_keyword ++ O.txt " "
                else O.noop)
@@ -713,9 +734,23 @@ module Make (Syntax : SYNTAX) = struct
 
     let type_decl ?(is_substitution = false)
         ((recursive, t) : Lang.Signature.recursive * Lang.TypeDecl.t) =
+      let keyword' =
+        match recursive with
+        | Ordinary | Rec -> O.keyword "type"
+        | And -> O.keyword "and"
+        | Nonrec -> O.keyword "type" ++ O.txt " " ++ O.keyword "nonrec"
+      in
       let tyname = Paths.Identifier.name t.id in
+      let tconstr =
+        match t.equation.params with
+        | [] -> O.txt tyname
+        | l ->
+            let params = format_params l in
+            Syntax.Type.handle_constructor_params (O.txt tyname) params
+      in
+      let intro = keyword' ++ O.txt " " ++ tconstr in
       let constraints = format_constraints t.equation.constraints in
-      let manifest, need_private =
+      let manifest, need_private, long_prefix =
         match t.equation.manifest with
         | Some (Odoc_model.Lang.TypeExpr.Polymorphic_variant variant) ->
             let code =
@@ -725,19 +760,22 @@ module Make (Syntax : SYNTAX) = struct
             in
             let manifest =
               O.documentedSrc
-                (O.txt (if is_substitution then " := " else " = ")
+                (O.ignore intro
+                ++ O.txt (if is_substitution then " :=" else " =")
+                ++ O.sp
                 ++
                 if t.equation.private_ then
                   O.keyword Syntax.Type.private_keyword ++ O.txt " "
                 else O.noop)
               @ code
             in
-            (manifest, false)
+            (manifest, false, O.noop)
         | _ ->
             let manifest, need_private =
               format_manifest ~is_substitution t.equation
             in
-            (O.documentedSrc manifest, need_private)
+            let text = O.ignore intro ++ manifest in
+            (O.documentedSrc @@ text, need_private, text)
       in
       let representation =
         match t.representation with
@@ -751,7 +789,7 @@ module Make (Syntax : SYNTAX) = struct
             in
             if List.length content > 0 then
               O.documentedSrc
-                (O.txt " = "
+                (O.ignore long_prefix ++ O.txt " =" ++ O.sp
                 ++
                 if need_private then
                   O.keyword Syntax.Type.private_keyword ++ O.txt " "
@@ -759,22 +797,8 @@ module Make (Syntax : SYNTAX) = struct
               @ content
             else []
       in
-      let tconstr =
-        match t.equation.params with
-        | [] -> O.txt tyname
-        | l ->
-            let params = format_params l in
-            Syntax.Type.handle_constructor_params (O.txt tyname) params
-      in
       let content =
-        let keyword' =
-          match recursive with
-          | Ordinary | Rec -> O.keyword "type"
-          | And -> O.keyword "and"
-          | Nonrec -> O.keyword "type" ++ O.txt " " ++ O.keyword "nonrec"
-        in
-        O.documentedSrc (keyword' ++ O.txt " " ++ tconstr)
-        @ manifest @ representation
+        O.documentedSrc intro @ manifest @ representation
         @ O.documentedSrc constraints
         @ O.documentedSrc
             (if Syntax.Type.type_def_semicolon then O.txt ";" else O.noop)
@@ -799,11 +823,12 @@ module Make (Syntax : SYNTAX) = struct
       let name = Paths.Identifier.name t.id in
       let content =
         O.documentedSrc
-          (O.keyword Syntax.Value.variable_keyword
-          ++ O.txt " " ++ O.txt name
-          ++ O.txt Syntax.Type.annotation_separator
-          ++ type_expr t.type_
-          ++ if semicolon then O.txt ";" else O.noop)
+          (O.box_hv
+          @@ O.keyword Syntax.Value.variable_keyword
+             ++ O.txt " " ++ O.txt name
+             ++ O.txt Syntax.Type.annotation_separator
+             ++ O.cut ++ type_expr t.type_
+             ++ if semicolon then O.txt ";" else O.noop)
       in
       let attr = [ "value" ] @ extra_attr in
       let anchor = path_to_id t.id in
@@ -1129,7 +1154,9 @@ module Make (Syntax : SYNTAX) = struct
             | TypExt e -> continue @@ extension e
             | Exception e -> continue @@ exn e
             | Value v -> continue @@ value v
-            | Open _ -> loop rest acc_items
+            | Open o ->
+                let items = Sectioning.comment_items o.doc in
+                loop rest (List.rev_append items acc_items)
             | Comment `Stop ->
                 let rest =
                   Utils.skip_until rest ~p:(function
@@ -1188,7 +1215,8 @@ module Make (Syntax : SYNTAX) = struct
       let path = Link.from_path (t.manifest :> Paths.Path.t) in
       let content =
         O.documentedSrc
-          (O.keyword "module" ++ O.txt " " ++ O.txt name ++ O.txt " := " ++ path)
+          (O.keyword "module" ++ O.txt " " ++ O.txt name ++ O.txt " :=" ++ O.sp
+         ++ path)
       in
       let attr = [ "module-substitution" ] in
       let anchor = path_to_id t.id in
@@ -1197,14 +1225,17 @@ module Make (Syntax : SYNTAX) = struct
 
     and module_type_substitution (t : Odoc_model.Lang.ModuleTypeSubstitution.t)
         =
+      let prefix =
+        O.keyword "module" ++ O.txt " " ++ O.keyword "type" ++ O.txt " "
+      in
       let modname = Paths.Identifier.name t.id in
       let modname, expansion_doc, mty =
         module_type_manifest ~subst:true modname t.id t.doc (Some t.manifest)
+          prefix
       in
       let content =
-        O.documentedSrc
-          (O.keyword "module" ++ O.txt " " ++ O.keyword "type" ++ O.txt " ")
-        @ modname @ mty
+        O.documentedSrc (prefix ++ modname)
+        @ mty
         @ O.documentedSrc
             (if Syntax.Mod.close_tag_semicolon then O.txt ";" else O.noop)
       in
@@ -1298,7 +1329,7 @@ module Make (Syntax : SYNTAX) = struct
       in
       let modname, status, expansion, expansion_doc =
         match expansion with
-        | None -> (O.documentedSrc (O.txt modname), `Default, None, None)
+        | None -> (O.txt modname, `Default, None, None)
         | Some (expansion_doc, items) ->
             let status =
               match t.type_ with
@@ -1311,17 +1342,17 @@ module Make (Syntax : SYNTAX) = struct
               make_expansion_page modname `Mod url [ t.doc; expansion_doc ]
                 items
             in
-            (O.documentedSrc link, status, Some page, Some expansion_doc)
+            (link, status, Some page, Some expansion_doc)
       in
-      let summary = mdexpr_in_decl t.id t.type_ in
+      let intro = O.keyword "module" ++ O.txt " " ++ modname in
+      let summary = O.ignore intro ++ mdexpr_in_decl t.id t.type_ in
       let modexpr =
         attach_expansion ~status
           (Syntax.Type.annotation_separator, "sig", "end")
           expansion summary
       in
       let content =
-        O.documentedSrc (O.keyword "module" ++ O.txt " ")
-        @ modname @ modexpr
+        O.documentedSrc intro @ modexpr
         @ O.documentedSrc
             (if Syntax.Mod.close_tag_semicolon then O.txt ";" else O.noop)
       in
@@ -1341,12 +1372,12 @@ module Make (Syntax : SYNTAX) = struct
     and mdexpr_in_decl (base : Paths.Identifier.Module.t) md =
       let sig_dotdotdot =
         O.txt Syntax.Type.annotation_separator
-        ++ Syntax.Mod.open_tag ++ O.txt " ... " ++ Syntax.Mod.close_tag
+        ++ O.cut ++ Syntax.Mod.open_tag ++ O.txt " ... " ++ Syntax.Mod.close_tag
       in
       match md with
       | Alias (_, Some se) -> simple_expansion_in_decl base se
       | Alias (p, _) when not Paths.Path.(is_hidden (p :> t)) ->
-          O.txt " = " ++ mdexpr md
+          O.txt " =" ++ O.sp ++ mdexpr md
       | Alias _ -> sig_dotdotdot
       | ModuleType mt -> mty_in_decl (base :> Paths.Identifier.Signature.t) mt
 
@@ -1354,7 +1385,7 @@ module Make (Syntax : SYNTAX) = struct
       | Alias (mod_path, _) -> Link.from_path (mod_path :> Paths.Path.t)
       | ModuleType mt -> mty mt
 
-    and module_type_manifest ~subst modname id doc manifest =
+    and module_type_manifest ~subst modname id doc manifest prefix =
       let expansion =
         match manifest with
         | None -> None
@@ -1362,33 +1393,38 @@ module Make (Syntax : SYNTAX) = struct
       in
       let modname, expansion, expansion_doc =
         match expansion with
-        | None -> (O.documentedSrc @@ O.txt modname, None, None)
+        | None -> (O.txt modname, None, None)
         | Some (expansion_doc, items) ->
             let url = Url.Path.from_identifier id in
             let link = path url [ inline @@ Text modname ] in
             let page =
               make_expansion_page modname `Mty url [ doc; expansion_doc ] items
             in
-            (O.documentedSrc link, Some page, Some expansion_doc)
+            (link, Some page, Some expansion_doc)
       in
       let summary =
         match manifest with
         | None -> O.noop
-        | Some expr -> (if subst then O.txt " := " else O.txt " = ") ++ mty expr
+        | Some expr ->
+            O.ignore (prefix ++ modname)
+            ++ (if subst then O.txt " :=" ++ O.sp else O.txt " =" ++ O.sp)
+            ++ mty expr
       in
       ( modname,
         expansion_doc,
         attach_expansion (" = ", "sig", "end") expansion summary )
 
     and module_type (t : Odoc_model.Lang.ModuleType.t) =
+      let prefix =
+        O.keyword "module" ++ O.txt " " ++ O.keyword "type" ++ O.txt " "
+      in
       let modname = Paths.Identifier.name t.id in
       let modname, expansion_doc, mty =
-        module_type_manifest ~subst:false modname t.id t.doc t.expr
+        module_type_manifest ~subst:false modname t.id t.doc t.expr prefix
       in
       let content =
-        O.documentedSrc
-          (O.keyword "module" ++ O.txt " " ++ O.keyword "type" ++ O.txt " ")
-        @ modname @ mty
+        O.documentedSrc (prefix ++ modname)
+        @ mty
         @ O.documentedSrc
             (if Syntax.Mod.close_tag_semicolon then O.txt ";" else O.noop)
       in
@@ -1414,9 +1450,9 @@ module Make (Syntax : SYNTAX) = struct
       | _ -> false
 
     and mty_with subs expr =
-      umty expr ++ O.txt " " ++ O.keyword "with" ++ O.txt " "
+      umty expr ++ O.sp ++ O.keyword "with" ++ O.txt " "
       ++ O.list
-           ~sep:(O.txt " " ++ O.keyword "and" ++ O.txt " ")
+           ~sep:(O.cut ++ O.txt " " ++ O.keyword "and" ++ O.txt " ")
            ~f:(fun x -> O.span (substitution x))
            subs
 
@@ -1462,7 +1498,7 @@ module Make (Syntax : SYNTAX) = struct
         | Functor (Unit, expr) ->
             (if Syntax.Mod.functor_keyword then O.keyword "functor" else O.noop)
             ++ O.span (O.txt " () " ++ Syntax.Type.arrow)
-            ++ O.txt " " ++ mty expr
+            ++ O.sp ++ mty expr
         | Functor (Named arg, expr) ->
             let arg_expr = arg.expr in
             let stop_before = expansion_of_module_type_expr arg_expr = None in
@@ -1476,15 +1512,16 @@ module Make (Syntax : SYNTAX) = struct
               | Ok href -> resolved href [ inline @@ Text name ]
             in
             (if Syntax.Mod.functor_keyword then O.keyword "functor" else O.noop)
-            ++ O.span
-                 (O.txt " (" ++ name
-                 ++ O.txt Syntax.Type.annotation_separator
-                 ++ mty arg_expr ++ O.txt ")" ++ O.txt " " ++ Syntax.Type.arrow
-                 )
-            ++ O.txt " " ++ mty expr
+            ++ (O.box_hv @@ O.span
+               @@ O.txt " (" ++ name
+                  ++ O.txt Syntax.Type.annotation_separator
+                  ++ mty arg_expr ++ O.txt ")" ++ O.txt " " ++ Syntax.Type.arrow
+               )
+            ++ O.sp ++ mty expr
         | With { w_expr; _ } when is_elidable_with_u w_expr ->
             Syntax.Mod.open_tag ++ O.txt " ... " ++ Syntax.Mod.close_tag
-        | With { w_substitutions; w_expr; _ } -> mty_with w_substitutions w_expr
+        | With { w_substitutions; w_expr; _ } ->
+            O.box_hv @@ mty_with w_substitutions w_expr
         | TypeOf { t_desc; _ } -> mty_typeof t_desc
         | Signature _ ->
             Syntax.Mod.open_tag ++ O.txt " ... " ++ Syntax.Mod.close_tag
@@ -1494,9 +1531,9 @@ module Make (Syntax : SYNTAX) = struct
         =
      fun base -> function
       | (Path _ | Signature _ | With _ | TypeOf _) as m ->
-          O.txt Syntax.Type.annotation_separator ++ mty m
+          O.txt Syntax.Type.annotation_separator ++ O.cut ++ mty m
       | Functor _ as m when not Syntax.Mod.functor_contraction ->
-          O.txt Syntax.Type.annotation_separator ++ mty m
+          O.txt Syntax.Type.annotation_separator ++ O.cut ++ mty m
       | Functor (arg, expr) ->
           let text_arg =
             match arg with
@@ -1516,11 +1553,12 @@ module Make (Syntax : SYNTAX) = struct
                   | Error _ -> O.txt name
                   | Ok href -> resolved href [ inline @@ Text name ]
                 in
-                O.txt "(" ++ name
-                ++ O.txt Syntax.Type.annotation_separator
-                ++ mty arg.expr ++ O.txt ")"
+                O.box_hv
+                @@ O.txt "(" ++ name
+                   ++ O.txt Syntax.Type.annotation_separator
+                   ++ O.cut ++ mty arg.expr ++ O.txt ")"
           in
-          O.txt " " ++ text_arg ++ mty_in_decl base expr
+          O.sp ++ text_arg ++ mty_in_decl base expr
 
     (* TODO : Centralize the list juggling for type parameters *)
     and type_expr_in_subst td typath =
@@ -1532,35 +1570,42 @@ module Make (Syntax : SYNTAX) = struct
     and substitution : Odoc_model.Lang.ModuleType.substitution -> text =
       function
       | ModuleEq (frag_mod, md) ->
-          O.keyword "module" ++ O.sp
-          ++ Link.from_fragment (frag_mod :> Paths.Fragment.leaf)
-          ++ O.sp ++ O.txt "= " ++ mdexpr md
+          O.box_hv
+          @@ O.keyword "module" ++ O.txt " "
+             ++ Link.from_fragment (frag_mod :> Paths.Fragment.leaf)
+             ++ O.txt " =" ++ O.sp ++ mdexpr md
       | ModuleTypeEq (frag_mty, md) ->
-          O.keyword "module" ++ O.txt " " ++ O.keyword "type" ++ O.txt " "
-          ++ Link.from_fragment (frag_mty :> Paths.Fragment.leaf)
-          ++ O.txt " = " ++ mty md
+          O.box_hv
+          @@ O.keyword "module" ++ O.txt " " ++ O.keyword "type" ++ O.txt " "
+             ++ Link.from_fragment (frag_mty :> Paths.Fragment.leaf)
+             ++ O.txt " =" ++ O.sp ++ mty md
       | TypeEq (frag_typ, td) ->
-          O.keyword "type" ++ O.sp
-          ++ type_expr_in_subst td (frag_typ :> Paths.Fragment.leaf)
-          ++ fst (format_manifest td)
-          ++ format_constraints td.Odoc_model.Lang.TypeDecl.Equation.constraints
+          O.box_hv
+          @@ O.keyword "type" ++ O.txt " "
+             ++ type_expr_in_subst td (frag_typ :> Paths.Fragment.leaf)
+             ++ fst (format_manifest td)
+             ++ format_constraints
+                  td.Odoc_model.Lang.TypeDecl.Equation.constraints
       | ModuleSubst (frag_mod, mod_path) ->
-          O.keyword "module" ++ O.sp
-          ++ Link.from_fragment (frag_mod :> Paths.Fragment.leaf)
-          ++ O.sp ++ O.txt ":= "
-          ++ Link.from_path (mod_path :> Paths.Path.t)
+          O.box_hv
+          @@ O.keyword "module" ++ O.txt " "
+             ++ Link.from_fragment (frag_mod :> Paths.Fragment.leaf)
+             ++ O.txt " :=" ++ O.sp
+             ++ Link.from_path (mod_path :> Paths.Path.t)
       | ModuleTypeSubst (frag_mty, md) ->
-          O.keyword "module" ++ O.txt " " ++ O.keyword "type" ++ O.txt " "
-          ++ Link.from_fragment (frag_mty :> Paths.Fragment.leaf)
-          ++ O.txt " := " ++ mty md
+          O.box_hv
+          @@ O.keyword "module" ++ O.txt " " ++ O.keyword "type" ++ O.txt " "
+             ++ Link.from_fragment (frag_mty :> Paths.Fragment.leaf)
+             ++ O.txt " :=" ++ O.sp ++ mty md
       | TypeSubst (frag_typ, td) -> (
-          O.keyword "type" ++ O.sp
-          ++ type_expr_in_subst td (frag_typ :> Paths.Fragment.leaf)
-          ++ O.sp ++ O.txt ":= "
-          ++
-          match td.Lang.TypeDecl.Equation.manifest with
-          | None -> assert false (* cf loader/cmti *)
-          | Some te -> type_expr te)
+          O.box_hv
+          @@ O.keyword "type" ++ O.txt " "
+             ++ type_expr_in_subst td (frag_typ :> Paths.Fragment.leaf)
+             ++ O.txt " :=" ++ O.sp
+             ++
+             match td.Lang.TypeDecl.Equation.manifest with
+             | None -> assert false (* cf loader/cmti *)
+             | Some te -> type_expr te)
 
     and include_ (t : Odoc_model.Lang.Include.t) =
       let decl_hidden =
