@@ -510,11 +510,7 @@ and read_structure_item env parent item =
     | Tstr_modtype mtd ->
         [ModuleType (Cmti.read_module_type_declaration env parent mtd)]
     | Tstr_open o ->
-#if OCAML_VERSION < (4,8,0)
-        ignore(o); []
-#else
         [Open (read_open env parent o)]
-#endif
     | Tstr_include incl ->
         read_include env parent incl
     | Tstr_class cls ->
@@ -566,11 +562,16 @@ and read_include env parent incl =
   | _ ->
     content.items
 
-#if OCAML_VERSION >= (4,8,0)
 and read_open env parent o =
-  let expansion, _ = Cmi.read_signature_noenv env parent (Odoc_model.Compat.signature o.open_bound_items) in
-  Open.{expansion}
-#endif
+  let container = (parent : Identifier.Signature.t :> Identifier.LabelParent.t) in
+  let doc = Doc_attr.attached_no_tag container o.open_attributes in
+  #if OCAML_VERSION >= (4,8,0)
+  let signature = o.open_bound_items in
+  #else
+  let signature = [] in
+  #endif
+  let expansion, _ = Cmi.read_signature_noenv env parent (Odoc_model.Compat.signature signature) in
+  Open.{expansion; doc}
 
 and read_structure :
       'tags. 'tags Odoc_model.Semantics.handle_internal_tags -> _ -> _ -> _ ->
