@@ -4,9 +4,9 @@
    output the result to. *)
 
 open Odoc_odoc
-open Cmdliner
+open Compatcmdliner
 
-let convert_syntax : Odoc_document.Renderer.syntax Arg.converter =
+let convert_syntax : Odoc_document.Renderer.syntax Arg.converter2 =
   let syntax_parser str =
     match str with
     | "ml" | "ocaml" -> `Ok Odoc_document.Renderer.OCaml
@@ -18,7 +18,7 @@ let convert_syntax : Odoc_document.Renderer.syntax Arg.converter =
   in
   (syntax_parser, syntax_printer)
 
-let convert_directory ?(create = false) () : Fs.Directory.t Arg.converter =
+let convert_directory ?(create = false) () : Fs.Directory.t Arg.converter2 =
   let dir_parser, dir_printer = Arg.string in
   let odoc_dir_parser str =
     let () = if create then Fs.Directory.(mkdir_p (of_string str)) in
@@ -110,7 +110,7 @@ module Compile : sig
 
   val cmd : unit Term.t
 
-  val info : Term.info
+  val info : Term.info2
 end = struct
   let has_page_prefix file =
     file |> Fs.File.basename |> Fs.File.to_string
@@ -250,7 +250,7 @@ end
 module Odoc_link : sig
   val cmd : unit Term.t
 
-  val info : Term.info
+  val info : Term.info2
 end = struct
   let get_output_file ~output_file ~input =
     match output_file with
@@ -297,11 +297,11 @@ module type S = sig
 end
 
 module Make_renderer (R : S) : sig
-  val process : unit Term.t * Term.info
+  val process : unit Term.t * Term.info2
 
-  val targets : unit Term.t * Term.info
+  val targets : unit Term.t * Term.info2
 
-  val generate : unit Term.t * Term.info
+  val generate : unit Term.t * Term.info2
 end = struct
   let input =
     let doc = "Input file" in
@@ -403,6 +403,55 @@ end = struct
   let targets = Targets.(cmd, info)
 end
 
+module Odoc_html_url : sig
+  val cmd : unit Term.t
+
+  val info : Term.info2
+end = struct
+  let root_url =
+    let doc =
+      "A string to prepend to the generated relative url. A separating / is \
+       added if needed."
+    in
+    Arg.(value & opt (some string) None & info [ "r"; "root-url" ] ~doc)
+
+  let reference =
+    let doc = "The reference to be resolved and whose url to be generated." in
+    Arg.(required & pos 0 (some string) None & info ~doc ~docv:"REF" [])
+
+  let reference_to_url = Url.reference_to_url_html
+
+  let cmd =
+    Term.(
+      const handle_error
+      $ (const reference_to_url $ root_url $ odoc_file_directories $ reference))
+
+  let info =
+    Term.info ~doc:"Resolve a reference and output its corresponding url"
+      "html-url"
+end
+
+module Odoc_latex_url : sig
+  val cmd : unit Term.t
+
+  val info : Term.info2
+end = struct
+  let reference =
+    let doc = "The reference to be resolved and whose url to be generated." in
+    Arg.(required & pos 0 (some string) None & info ~doc ~docv:"REF" [])
+
+  let reference_to_url = Url.reference_to_url_latex
+
+  let cmd =
+    Term.(
+      const handle_error
+      $ (const reference_to_url $ odoc_file_directories $ reference))
+
+  let info =
+    Term.info ~doc:"Resolve a reference and output its corresponding url"
+      "latex-url"
+end
+
 module Odoc_html = Make_renderer (struct
   type args = Html_page.args
 
@@ -424,7 +473,7 @@ module Odoc_html = Make_renderer (struct
     Arg.(value & flag (info ~doc [ "indent" ]))
 
   (* Very basic validation and normalization for URI paths. *)
-  let convert_uri : Odoc_html.Tree.uri Arg.converter =
+  let convert_uri : Odoc_html.Tree.uri Arg.converter2 =
     let parser str =
       if String.length str = 0 then `Error "invalid URI"
       else
@@ -504,7 +553,7 @@ end)
 module Html_fragment : sig
   val cmd : unit Term.t
 
-  val info : Term.info
+  val info : Term.info2
 end = struct
   let html_fragment directories xref_base_uri output_file input_file
       warnings_options =
@@ -736,6 +785,8 @@ let () =
       Targets.Support_files.(cmd, info);
       Odoc_link.(cmd, info);
       Odoc_error.(cmd, info);
+      Odoc_html_url.(cmd, info);
+      Odoc_latex_url.(cmd, info);
     ]
   in
   let default =
@@ -748,7 +799,7 @@ let () =
         (String.concat ~sep:", " available_subcommands)
     in
     ( Term.(const print_default $ const ()),
-      Term.info ~version:"2.0.2" "odoc" )
+      Term.info ~version:"2.1.1" "odoc" )
   in
   match Term.eval_choice ~err:Format.err_formatter default subcommands with
   | `Error _ ->
