@@ -299,7 +299,7 @@ and Signature : sig
 end
 
 and Open : sig
-  type t = { expansion : Signature.t }
+  type t = { expansion : Signature.t; doc : CComment.docs }
 end
 
 and Include : sig
@@ -609,7 +609,7 @@ end
 module Of_Lang : sig
   type map
 
-  val empty : map
+  val empty : unit -> map
 
   val identifier :
     ('a -> 'b -> 'c) -> 'b -> 'a -> [> `Identifier of 'a | `Local of 'c ]
