@@ -88,115 +88,6 @@ end =
 
 include Cpath
 
-type local_path_error =
-  | ErrModule of module_
-  | ErrModuleType of module_type
-  | ErrType of type_
-
-exception LocalPath of local_path_error
-
-let rec resolved_module_path_of_cpath :
-    Resolved.module_ -> Path.Resolved.Module.t = function
-  | `Local _ as y -> raise (LocalPath (ErrModule (`Resolved y)))
-  | `Substituted y -> resolved_module_path_of_cpath y
-  | `Identifier (#Identifier.Path.Module.t as x) -> `Identifier x
-  | `Subst (a, b) ->
-      `Subst
-        (resolved_module_type_path_of_cpath a, resolved_module_path_of_cpath b)
-  | `Hidden x -> `Hidden (resolved_module_path_of_cpath x)
-  | `Canonical (a, b) ->
-      `Canonical (resolved_module_path_of_cpath a, module_path_of_cpath b)
-  | `Apply (a, b) ->
-      `Apply (resolved_module_path_of_cpath a, resolved_module_path_of_cpath b)
-  | `Alias (a, b) ->
-      `Alias (resolved_module_path_of_cpath a, resolved_module_path_of_cpath b)
-  | `Module (p, m) -> `Module (resolved_module_path_of_cpath_parent p, m)
-  | `OpaqueModule m -> `OpaqueModule (resolved_module_path_of_cpath m)
-
-and resolved_module_path_of_cpath_parent :
-    Resolved.parent -> Path.Resolved.Module.t = function
-  | `Module m -> resolved_module_path_of_cpath m
-  | `ModuleType _ | `FragmentRoot -> failwith "Can't do it"
-
-and resolved_module_type_path_of_cpath :
-    Resolved.module_type -> Path.Resolved.ModuleType.t = function
-  | `Local _ as y -> raise (LocalPath (ErrModuleType (`Resolved y)))
-  | `Identifier (#Identifier.ModuleType.t as x) -> `Identifier x
-  | `Substituted y -> resolved_module_type_path_of_cpath y
-  | `ModuleType (p, m) -> `ModuleType (resolved_module_path_of_cpath_parent p, m)
-  | `SubstT (p1, p2) ->
-      `SubstT
-        ( resolved_module_type_path_of_cpath p1,
-          resolved_module_type_path_of_cpath p2 )
-  | `AliasModuleType (m1, m2) ->
-      `AliasModuleType
-        ( resolved_module_type_path_of_cpath m1,
-          resolved_module_type_path_of_cpath m2 )
-  | `CanonicalModuleType (p1, p2) ->
-      `CanonicalModuleType
-        (resolved_module_type_path_of_cpath p1, module_type_path_of_cpath p2)
-  | `OpaqueModuleType m ->
-      `OpaqueModuleType (resolved_module_type_path_of_cpath m)
-
-and resolved_type_path_of_cpath : Resolved.type_ -> Path.Resolved.Type.t =
-  function
-  | `Identifier (#Odoc_model.Paths.Identifier.Path.Type.t as x) -> `Identifier x
-  | `Local _ as y -> raise (LocalPath (ErrType (`Resolved y)))
-  | `Substituted y -> resolved_type_path_of_cpath y
-  | `CanonicalType (t1, t2) ->
-      `CanonicalType (resolved_type_path_of_cpath t1, type_path_of_cpath t2)
-  | `Type (p, m) -> `Type (resolved_module_path_of_cpath_parent p, m)
-  | `Class (p, m) -> `Class (resolved_module_path_of_cpath_parent p, m)
-  | `ClassType (p, m) -> `ClassType (resolved_module_path_of_cpath_parent p, m)
-
-and resolved_class_type_path_of_cpath :
-    Resolved.class_type -> Path.Resolved.ClassType.t = function
-  | `Identifier (#Odoc_model.Paths.Identifier.Path.ClassType.t as x) ->
-      `Identifier x
-  | `Local ident ->
-      raise
-        (LocalPath (ErrType (`Resolved (`Local (ident :> Ident.path_type)))))
-  | `Substituted y -> resolved_class_type_path_of_cpath y
-  | `Class (p, m) -> `Class (resolved_module_path_of_cpath_parent p, m)
-  | `ClassType (p, m) -> `ClassType (resolved_module_path_of_cpath_parent p, m)
-
-and module_path_of_cpath : module_ -> Path.Module.t = function
-  | `Resolved r -> `Resolved (resolved_module_path_of_cpath r)
-  | `Dot (p, x) -> `Dot (module_path_of_cpath p, x)
-  | `Module (_p, _n) -> failwith "Probably shouldn't happen"
-  | `Identifier (x, b) -> `Identifier (x, b)
-  | `Local _ as y -> raise (LocalPath (ErrModule y))
-  | `Substituted p -> module_path_of_cpath p
-  | `Root x -> `Root x
-  | `Forward x -> `Forward x
-  | `Apply (m1, m2) -> `Apply (module_path_of_cpath m1, module_path_of_cpath m2)
-
-and module_type_path_of_cpath : module_type -> Path.ModuleType.t = function
-  | `Resolved r -> `Resolved (resolved_module_type_path_of_cpath r)
-  | `Identifier (x, b) -> `Identifier (x, b)
-  | `Local _ as y -> raise (LocalPath (ErrModuleType y))
-  | `Substituted r -> module_type_path_of_cpath r
-  | `Dot (p, x) -> `Dot (module_path_of_cpath p, x)
-  | `ModuleType (_p, _n) -> failwith "Probably shouldn't happen"
-
-and type_path_of_cpath : type_ -> Path.Type.t = function
-  | `Resolved r -> `Resolved (resolved_type_path_of_cpath r)
-  | `Identifier (x, b) -> `Identifier (x, b)
-  | `Local _ as y -> raise (LocalPath (ErrType y))
-  | `Substituted r -> type_path_of_cpath r
-  | `Dot (p, x) -> `Dot (module_path_of_cpath p, x)
-  | `Type (_, _) | `Class (_, _) | `ClassType (_, _) ->
-      failwith "Probably shouldn't happen"
-
-and class_type_path_of_cpath : class_type -> Path.ClassType.t = function
-  | `Resolved r -> `Resolved (resolved_class_type_path_of_cpath r)
-  | `Identifier (x, b) -> `Identifier (x, b)
-  | `Local (ident, b) ->
-      raise (LocalPath (ErrType (`Local ((ident :> Ident.path_type), b))))
-  | `Substituted r -> class_type_path_of_cpath r
-  | `Dot (p, x) -> `Dot (module_path_of_cpath p, x)
-  | `Class (_, _) | `ClassType (_, _) -> failwith "Probably shouldn't happen"
-
 let rec is_resolved_module_substituted : Resolved.module_ -> bool = function
   | `Local _ -> false
   | `Substituted _ -> true
