@@ -61,7 +61,13 @@ let pp_lookup_type_list fmt ls =
   in
   Format.fprintf fmt "[%a]" inner ls
 
-type recorder = { mutable lookups : lookup_type list }
+module LookupTypeSet = Set.Make (struct
+  type t = lookup_type
+
+  let compare = compare
+end)
+
+type recorder = { mutable lookups : LookupTypeSet.t }
 
 module Maps = Odoc_model.Paths.Identifier.Maps
 module StringMap = Map.Make (String)
@@ -181,11 +187,11 @@ let has_resolver t = match t.resolver with None -> false | _ -> true
 let id t = t.id
 
 let with_recorded_lookups env f =
-  let recorder = { lookups = [] } in
+  let recorder = { lookups = LookupTypeSet.empty } in
   let env' = { env with recorder = Some recorder } in
   let restore () =
     match env.recorder with
-    | Some r -> r.lookups <- recorder.lookups @ r.lookups
+    | Some r -> r.lookups <- LookupTypeSet.union recorder.lookups r.lookups
     | None -> ()
   in
   try
@@ -222,7 +228,7 @@ let add_to_elts kind identifier component env =
     let other = ElementsById.find_by_id identifier env.ids in
     match other with
     | Some _ ->
-        (* Format.eprintf "Overriding duplicate env entry\n%!" *)
+        (* Format.eprintf "Overriding duplicate env entry: %s\n%!" (Identifier.name identifier); *)
         ()
     | None -> ()
   in
@@ -387,7 +393,7 @@ let module_of_unit : Odoc_model.Lang.Compilation_unit.t -> Component.Module.t =
             hidden = unit.hidden;
           }
       in
-      let ty = Component.Of_Lang.(module_ empty m) in
+      let ty = Component.Of_Lang.(module_ (empty ()) m) in
       ty
   | Pack _p ->
       let m =
@@ -401,7 +407,7 @@ let module_of_unit : Odoc_model.Lang.Compilation_unit.t -> Component.Module.t =
             hidden = unit.hidden;
           }
       in
-      let ty = Component.Of_Lang.(module_ empty m) in
+      let ty = Component.Of_Lang.(module_ (empty ()) m) in
       ty
 
 let lookup_root_module name env =
@@ -419,10 +425,15 @@ let lookup_root_module name env =
   in
   (match (env.recorder, result) with
   | Some r, Some Forward ->
-      r.lookups <- RootModule (name, Some `Forward) :: r.lookups
+      r.lookups <-
+        LookupTypeSet.add (RootModule (name, Some `Forward)) r.lookups
   | Some r, Some (Resolved (root, _, _)) ->
-      r.lookups <- RootModule (name, Some (`Resolved root.digest)) :: r.lookups
-  | Some r, None -> r.lookups <- RootModule (name, None) :: r.lookups
+      r.lookups <-
+        LookupTypeSet.add
+          (RootModule (name, Some (`Resolved root.digest)))
+          r.lookups
+  | Some r, None ->
+      r.lookups <- LookupTypeSet.add (RootModule (name, None)) r.lookups
   | None, _ -> ());
   result
 
@@ -448,7 +459,8 @@ let lookup_by_name scope name env =
         List.iter
           (function
             | `Module (id, _) ->
-                r.lookups <- ModuleByName (name, id) :: r.lookups
+                r.lookups <-
+                  LookupTypeSet.add (ModuleByName (name, id)) r.lookups
             | _ -> ())
           (results :> Component.Element.any list)
     | None -> ()
@@ -475,8 +487,10 @@ let lookup_by_id (scope : 'a scope) id env : 'a option =
     match env.recorder with
     | Some r -> (
         match (result :> Component.Element.any) with
-        | `Module (id, _) -> r.lookups <- Module id :: r.lookups
-        | `ModuleType (id, _) -> r.lookups <- ModuleType id :: r.lookups
+        | `Module (id, _) ->
+            r.lookups <- LookupTypeSet.add (Module id) r.lookups
+        | `ModuleType (id, _) ->
+            r.lookups <- LookupTypeSet.add (ModuleType id) r.lookups
         | _ -> ())
     | None -> ()
   in
@@ -589,7 +603,7 @@ let n = ref 0
 let lookup_fragment_root env =
   let maybe_record_result res =
     match env.recorder with
-    | Some r -> r.lookups <- res :: r.lookups
+    | Some r -> r.lookups <- LookupTypeSet.add res r.lookups
     | None -> ()
   in
   match env.fragmentroot with
@@ -607,7 +621,8 @@ let add_functor_parameter : Odoc_model.Lang.FunctorParameter.t -> t -> t =
         Component.Module.
           {
             doc = [];
-            type_ = ModuleType Component.Of_Lang.(module_type_expr empty n.expr);
+            type_ =
+              ModuleType Component.Of_Lang.(module_type_expr (empty ()) n.expr);
             canonical = None;
             hidden = false;
           }
@@ -682,7 +697,7 @@ let open_class_signature : Odoc_model.Lang.ClassSignature.t -> t -> t =
       (fun env orig ->
         match orig with
         | Odoc_model.Lang.ClassSignature.Method m ->
-            let ty = method_ empty m in
+            let ty = method_ (empty ()) m in
             add_method m.Odoc_model.Lang.Method.id ty env
         | _ -> env)
       env s.items
@@ -691,71 +706,53 @@ let rec open_signature : Odoc_model.Lang.Signature.t -> t -> t =
   let open Component in
   let open Of_Lang in
   let module L = Odoc_model.Lang in
+  let ident_map = empty () in
   fun s e ->
     List.fold_left
       (fun env orig ->
         match ((orig : L.Signature.item), env.linking) with
         | Type (_, t), _ ->
-            let ty = type_decl empty t in
+            let ty = type_decl ident_map t in
             add_type t.L.TypeDecl.id ty env
         | Module (_, t), _ ->
-            let ty = Component.Delayed.put (fun () -> module_ empty t) in
+            let ty = Component.Delayed.put (fun () -> module_ ident_map t) in
             add_module
               (t.L.Module.id :> Identifier.Path.Module.t)
               ty
-              (docs empty t.L.Module.doc)
+              (docs ident_map t.L.Module.doc)
               env
         | ModuleType t, _ ->
-            let ty = module_type empty t in
+            let ty = module_type ident_map t in
             add_module_type t.L.ModuleType.id ty env
-        | ModuleTypeSubstitution t, _ ->
-            let ty =
-              module_type empty
-                {
-                  id = t.id;
-                  doc = t.doc;
-                  expr = Some t.manifest;
-                  canonical = None;
-                }
-            in
-            add_module_type t.L.ModuleTypeSubstitution.id ty env
-        | L.Signature.TypeSubstitution t, _ ->
-            let ty = type_decl empty t in
-            add_type t.L.TypeDecl.id ty env
-        | L.Signature.ModuleSubstitution m, _ ->
-            let _id = Ident.Of_Identifier.module_ m.id in
-            let doc = docs empty m.doc in
-            let ty =
-              Component.Delayed.put (fun () ->
-                  Of_Lang.(
-                    module_of_module_substitution
-                      (*                  { empty with modules = [ (m.id, id) ] } *)
-                      empty m))
-            in
-            add_module (m.id :> Identifier.Path.Module.t) ty doc env
+        | ModuleTypeSubstitution _, _
+        | L.Signature.TypeSubstitution _, _
+        | L.Signature.ModuleSubstitution _, _ ->
+            env
         | L.Signature.Class (_, c), _ ->
-            let ty = class_ empty c in
+            let ty = class_ ident_map c in
             add_class c.id ty env
         | L.Signature.ClassType (_, c), _ ->
-            let ty = class_type empty c in
+            let ty = class_type ident_map c in
             add_class_type c.id ty env
         | L.Signature.Include i, _ -> open_signature i.expansion.content env
-        | L.Signature.Open o, _ -> open_signature o.expansion env
+        | L.Signature.Open o, false -> open_signature o.expansion env
         (* The following are only added when linking *)
+        | L.Signature.Open o, true ->
+            add_comment (`Docs o.doc) (open_signature o.expansion env)
         | Comment c, true -> add_comment c env
         | TypExt te, true ->
-            let doc = docs empty te.doc in
+            let doc = docs ident_map te.doc in
             List.fold_left
               (fun env tec ->
-                let ty = extension_constructor empty tec in
+                let ty = extension_constructor ident_map tec in
                 add_extension_constructor tec.L.Extension.Constructor.id ty env)
               env te.L.Extension.constructors
             |> add_cdocs te.L.Extension.parent doc
         | Exception e, true ->
-            let ty = exception_ empty e in
+            let ty = exception_ ident_map e in
             add_exception e.L.Exception.id ty env
         | L.Signature.Value v, true ->
-            let ty = value empty v in
+            let ty = value ident_map v in
             add_value v.L.Value.id ty env
         (* Skip when compiling *)
         | Exception _, false -> env
@@ -764,6 +761,43 @@ let rec open_signature : Odoc_model.Lang.Signature.t -> t -> t =
         | L.Signature.Value _, false -> env)
       e s.items
 
+let open_type_substitution : Odoc_model.Lang.TypeDecl.t -> t -> t =
+ fun t env ->
+  let open Component in
+  let open Of_Lang in
+  let module L = Odoc_model.Lang in
+  let ty = type_decl (empty ()) t in
+  add_type t.L.TypeDecl.id ty env
+
+let open_module_substitution : Odoc_model.Lang.ModuleSubstitution.t -> t -> t =
+ fun m env ->
+  let open Component in
+  let open Of_Lang in
+  let module L = Odoc_model.Lang in
+  let _id = Ident.Of_Identifier.module_ m.id in
+  let doc = docs (empty ()) m.doc in
+  let ty =
+    Component.Delayed.put (fun () ->
+        Of_Lang.(
+          module_of_module_substitution
+            (*                  { empty with modules = [ (m.id, id) ] } *)
+            (empty ())
+            m))
+  in
+  add_module (m.id :> Identifier.Path.Module.t) ty doc env
+
+let open_module_type_substitution :
+    Odoc_model.Lang.ModuleTypeSubstitution.t -> t -> t =
+ fun t env ->
+  let open Component in
+  let open Of_Lang in
+  let module L = Odoc_model.Lang in
+  let ty =
+    module_type (empty ())
+      { id = t.id; doc = t.doc; expr = Some t.manifest; canonical = None }
+  in
+  add_module_type t.L.ModuleTypeSubstitution.id ty env
+
 let rec close_signature : Odoc_model.Lang.Signature.t -> t -> t =
   let module L = Odoc_model.Lang in
   fun s e ->
@@ -774,13 +808,13 @@ let rec close_signature : Odoc_model.Lang.Signature.t -> t -> t =
         | Type (_, t) -> remove t.L.TypeDecl.id env
         | Module (_, t) -> remove t.L.Module.id env
         | ModuleType t -> remove t.L.ModuleType.id env
-        | ModuleTypeSubstitution t -> remove t.L.ModuleTypeSubstitution.id env
-        | TypeSubstitution t -> remove t.L.TypeDecl.id env
-        | ModuleSubstitution t -> remove t.L.ModuleSubstitution.id env
         | Class (_, c) -> remove c.id env
         | ClassType (_, c) -> remove c.id env
         | Include i -> close_signature i.expansion.content env
         | Open o -> close_signature o.expansion env
+        | ModuleSubstitution _ | ModuleTypeSubstitution _ | TypeSubstitution _
+          ->
+            env
         (* The following are only added when linking *)
         | Exception _ -> env
         | TypExt _ -> env
@@ -818,6 +852,9 @@ let env_of_page page resolver =
   let initial_env = open_page page empty in
   set_resolver initial_env resolver |> open_units resolver
 
+let env_for_reference resolver =
+  set_resolver empty resolver |> open_units resolver
+
 let env_for_testing ~linking = { empty with linking }
 
 let verify_lookups env lookups =
@@ -868,10 +905,10 @@ let verify_lookups env lookups =
            true
        end*)
   in
-  let result = not (List.exists bad_lookup lookups) in
+  let result = not (LookupTypeSet.exists bad_lookup lookups) in
   (* If we're recording lookups, make sure it looks like we
       looked all this stuff up *)
   (match (result, env.recorder) with
-  | true, Some r -> r.lookups <- r.lookups @ lookups
+  | true, Some r -> r.lookups <- LookupTypeSet.union r.lookups lookups
   | _ -> ());
   result
